///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoQueryCache.cc
// -----------------
// Cego query cache implementation
//      
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2015 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoQueryCache
// 
// Description: 
//
//
///////////////////////////////////////////////////////////////////////////////

#include <lfcbase/ThreadLock.h>
#include "CegoQueryCache.h"
#include "CegoXMLdef.h"

#include <stdlib.h>
#include <string.h>

#ifndef _REENTRANT
#define _REENTRANT    /* basic 3-lines for threads */
#endif

static ThreadLock cacheLock("QUERYCACHE");
extern bool __lockStatOn;

CegoQueryCache::QueryCacheEntry::QueryCacheEntry()
{
    _pCacheArray = 0;
    _numHit=0;
}

CegoQueryCache::QueryCacheEntry::QueryCacheEntry(const Chain& queryId)
{
    _queryId = queryId;
    _pCacheArray = 0;
    _numHit=0;
}

CegoQueryCache::QueryCacheEntry::QueryCacheEntry(const Chain& queryId, const SetT<CegoObject>& tableList, ListT< ListT<CegoFieldValue> >* pCacheList, const ListT<CegoField>& cacheSchema)
{
    _queryId = queryId;
    _tableList = tableList;

    _pCacheArray = new CegoFieldValue**[ pCacheList->Size() ];
    ListT<CegoFieldValue>* pFVL = pCacheList->First();
    _numRow = 0;
    _numCol = 0;
    while ( pFVL )
    {
	CegoFieldValue** pCFVL = new CegoFieldValue*[pFVL->Size()];
	CegoFieldValue* pFV = pFVL->First();
	int i = 0;
	while ( pFV )
	{
	    CegoFieldValue* pCFV = new CegoFieldValue(pFV->getLocalCopy());
	    pCFVL[i] = pCFV;
	    pFV = pFVL->Next();
	    i++;
	}
	_numCol = i;
       
	_pCacheArray[_numRow] = pCFVL;
	_numRow++;
	pFVL = pCacheList->Next();	
    }
    
    _cacheSchema = cacheSchema;
    _numHit=1;
}

CegoQueryCache::QueryCacheEntry::~QueryCacheEntry()
{
}

void CegoQueryCache::QueryCacheEntry::cleanCache()
{
    for ( int i=0; i<_numRow; i++ )
    {
	for ( int j=0; j<_numCol; j++ )
	    delete _pCacheArray[i][j];
	delete _pCacheArray[i];
    }
    delete _pCacheArray;
    _pCacheArray = 0;
}


const int CegoQueryCache::QueryCacheEntry::getSize() const
{
    int s =  _queryId.length();
    CegoObject *pTD = _tableList.First();
    while ( pTD )
    {
	s += pTD->size();
	pTD = _tableList.Next();
    }

    CegoField *pF = _cacheSchema.First();
    while ( pF )
    {
	s += pF->size();
	pF = _cacheSchema.Next();
    }

    for ( int i=0; i<_numRow; i++ )
    {
	for ( int j=0; j<_numCol; j++ )
	{
	    CegoFieldValue *pF = _pCacheArray[i][j];
	    s += pF->size();
	}
    }

    return s;
	
}

const Chain CegoQueryCache::QueryCacheEntry::getQueryId() const
{
    return _queryId;
}

int CegoQueryCache::QueryCacheEntry::getNumRows() const
{
    return _numRow;
}

const SetT<CegoObject>& CegoQueryCache::QueryCacheEntry::getTableList() const
{
    return _tableList; 
}

CegoFieldValue*** CegoQueryCache::QueryCacheEntry::getCache()
{
    return _pCacheArray;
}

const ListT<CegoField>& CegoQueryCache::QueryCacheEntry::getSchema() const
{
    return _cacheSchema;
}

unsigned long CegoQueryCache::QueryCacheEntry::getHit() const
{
    return _numHit;
}

void CegoQueryCache::QueryCacheEntry::incHit()
{
    _numHit++;
}

CegoQueryCache::QueryCacheEntry& CegoQueryCache::QueryCacheEntry::operator = ( const CegoQueryCache::QueryCacheEntry& qce)
{
    _queryId = qce._queryId;
    _tableList = qce._tableList;
    _pCacheArray = qce._pCacheArray;
    _cacheSchema = qce._cacheSchema;
    _numHit = qce._numHit;
    _numRow = qce._numRow;
    _numCol = qce._numCol;
    return (*this);
}

bool CegoQueryCache::QueryCacheEntry::operator == ( const CegoQueryCache::QueryCacheEntry& qce)
{
    if ( _queryId == qce._queryId )
	return true;
    return false;
}

CegoQueryCache::CegoQueryCache(int maxEntry, int maxSize)
{
    cacheLock.init(LCKMNG_LOCKWAITDELAY, __lockStatOn);  
    _maxEntry = maxEntry;
    _maxSize = maxSize;
}

CegoQueryCache::~CegoQueryCache()
{
    clean();
}

void CegoQueryCache::PR()
{
    cacheLock.readLock(DBM_LOCKTIMEOUT);
}

void CegoQueryCache::PW()
{
    cacheLock.writeLock(DBM_LOCKTIMEOUT);
}

void CegoQueryCache::V()
{
    cacheLock.unlock();
}

Element* CegoQueryCache::getCacheInfo()
{
    Element* pCacheInfo = new Element(XML_CACHEINFO_ELEMENT);

    Element *pN = new Element(XML_CACHE_ELEMENT);
    pN->setAttribute(XML_ATTRNAME_ATTR, Chain("MaxEntry"));
    pN->setAttribute(XML_VALUE_ATTR, Chain(_maxEntry));
    pCacheInfo->addContent(pN);

    pN = new Element(XML_CACHE_ELEMENT);
    pN->setAttribute(XML_ATTRNAME_ATTR, Chain("MaxSize"));
    pN->setAttribute(XML_VALUE_ATTR, Chain(_maxSize));
    pCacheInfo->addContent(pN);
    
    return pCacheInfo;
}

Element* CegoQueryCache::getCacheList()
{
    Element* pCacheInfo = new Element(XML_CACHEINFO_ELEMENT);
    PR();

    QueryCacheEntry *pCE = _queryCache.First();	
    while ( pCE )
    {
	Element *pN = new Element(XML_CACHE_ELEMENT);
	pN->setAttribute(XML_ID_ATTR, pCE->getQueryId());
	pN->setAttribute(XML_NUMROWS_ATTR, Chain(pCE->getNumRows()));
	pN->setAttribute(XML_NUMHITS_ATTR, Chain(pCE->getHit()));
	pN->setAttribute(XML_SIZE_ATTR, Chain(pCE->getSize()));
	pCacheInfo->addContent(pN);
	pCE = _queryCache.Next();
    }
    V();
    return pCacheInfo;
}


int CegoQueryCache::getMaxEntry() const
{
    return _maxEntry;
}

void CegoQueryCache::setMaxEntry(int maxEntry)
{
    _maxEntry = maxEntry;
}

int CegoQueryCache::getMaxSize() const
{
    return _maxSize;
}

void CegoQueryCache::setMaxSize(int maxSize)
{
    _maxSize = maxSize;
}

const int CegoQueryCache::getNumQueryCache() const
{
    return _queryCache.Size();
}

void CegoQueryCache::invalidate(const CegoObject& td)
{
    
    PW();
    QueryCacheEntry *pQCE = _queryCache.First();
    while ( pQCE )
    {
	if ( pQCE->getTableList().Find(td) )
	{
	    pQCE->cleanCache();	
	    _queryCache.Remove(*pQCE);
	    pQCE = _queryCache.First();
	}
	else
	{
	    pQCE = _queryCache.Next();
	}
    }
    
    V();
    
}

CegoFieldValue*** CegoQueryCache::getEntry(const Chain& queryId, ListT<CegoField>& cacheSchema, int& cacheRows)
{
    CegoFieldValue*** pCE = 0;
    
    PR();
    QueryCacheEntry *pQCE = _queryCache.Find( QueryCacheEntry(queryId));
    if ( pQCE )
    {
	pQCE->incHit();
	pCE = pQCE->getCache();
	cacheSchema = pQCE->getSchema();
	cacheRows = pQCE->getNumRows();
    }
    V();
    
    return pCE;
}

void CegoQueryCache::addEntry(const Chain& queryId, const SetT<CegoObject>& tableList, ListT< ListT<CegoFieldValue> >* pCacheList, const ListT<CegoField>& cacheSchema )
{
    PW();

    QueryCacheEntry *pQCE = _queryCache.Find( QueryCacheEntry(queryId));
    if ( pQCE )
    {
	V();
	return;
    }
    if ( _queryCache.Size() > _maxEntry )
    {

	QueryCacheEntry *pRE = 0;
	unsigned long minHit = 0;
	QueryCacheEntry *pQCE = _queryCache.First();
	while ( pQCE )
	{
	    if ( minHit == 0 || pQCE->getHit() < minHit )
	    {
		pRE = pQCE;
		minHit = pQCE->getHit();
	    }
	    
	    pQCE = _queryCache.Next();
	}
	if ( pRE )
	{
	    pRE->cleanCache();	
	    _queryCache.Remove(*pRE);
	}
    }    
    _queryCache.Insert( QueryCacheEntry(queryId, tableList, pCacheList, cacheSchema) ); 
    V();
}

void CegoQueryCache::clean()
{
    PW();
    QueryCacheEntry *pCE = _queryCache.First();
    while ( pCE )
    {	
	pCE->cleanCache();	
	pCE = _queryCache.Next();
    }
    
    _queryCache.Empty();
    V();
}
